import os
from math import ceil
from tile import Tile
from PIL import Image

class Photo:
	'''
	Represents a photo to be tiled
	'''

	def __init__(self, path):
		self.binary = Image.open(path)
		self.dir = os.path.dirname(path)
		self.name, self.file_format = os.path.basename(path).split('.')
		self.width, self.height = self.binary.size

	def extract_tiles(self, tiling_level, tile_max_width, tile_max_height):
		'''
		Extract and save all possible tiles from the current resolution of the photo. No tile
		will have dimensions (width or height) bigger than 'tile_max_width' and 'tile_max_height'. All
		extrated tiles will be saved in a subdirectory named 'tiling_level'
		'''

		y = 0

		# while tile coordinate y is less than the maximum height range 
		while (y < (self.height / tile_max_height)):
			x = 0
			# while tile coordinate x is less than the maximum width range
			while (x < (self.width / tile_max_width)):
				# Define top-left and lower-right corner coordinates of the tile
				x_start = x * tile_max_width
				y_start = y * tile_max_height
				x_end = x_start + min(tile_max_width, (self.width - (x * tile_max_width)))
				y_end = y_start + min(tile_max_height, (self.height - (y * tile_max_height)))

				tile_boundaries = (x_start, y_start, x_end, y_end)

				# Crop the tile from the original photo and register its coordinates
				tile_binary = self.binary.crop(tile_boundaries)
				tile_coordinates = (x, y)

				# Create and save the tile to the file system
				Tile(tile_binary, tile_coordinates, self.file_format).save(os.path.join(self.dir, 'tiles', self.name, str(tiling_level)))

				x += 1
			y += 1

	def shrink(self, ratio):
		'''
		Scale photo resolution by 'ratio'
		'''

		scaled_width = ceil(self.width * ratio)
		scaled_height = ceil(self.height * ratio)

		self.binary = self.binary.resize((scaled_width, scaled_height))
		self.width = scaled_width
		self.height = scaled_height



import os
import shutil

def recursive_file_count(root_dir):
	'''
	Recursively counts how many files exist in 'root_dir' and all subdirectories
	'''

	return sum([len(files) for r, d, files in os.walk(root_dir)])

def recursive_dir_count(root_dir):
	'''
	Recursively counts how many directories exist in 'root_dir' and all subdirectories
	'''

	return sum([len(d) for r, d, files in os.walk(root_dir)])

def scan_dir(root_dir, include = 'dir'):
	'''
	Lists all files or directories exist in 'root_dir'. The distinction between what
	needs to be counted is made using the argument 'include'
	'''

	if include == 'dir':
		opt = 1
	elif include == 'file':
		opt = 2
	else:
		return []

	return [element for element in next(os.walk(root_dir))[opt]]

def create_dir(path):
	'''
	Creates a new directory as referenced by path
	'''

	os.mkdir(path)

def delete_dir(path):
	'''
	Deletes a directory referenced by path
	'''

	shutil.rmtree(path)



import os

class Tile:
	'''
	Represents a single tile of a photo
	'''

	def __init__(self, binary, coordinates, file_format):
		self.binary = binary
		self.width, self.height = self.binary.size
		self.coordinates = coordinates
		self.file_format = file_format

	def save(self, dest_directory):
		'''
		Saves a single tile to the destination directory 'dest_directory'
		'''

		tile_name = str(self.coordinates[0]) + '_' + str(self.coordinates[1]) + '.' + self.file_format
		file_path = os.path.join(dest_directory, tile_name)

		os.makedirs(dest_directory, exist_ok = True)
		
		self.binary.save(file_path, subsampling = 0, quality = 100)



import os
import re
import unittest
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import tilesman.tests.helpers.image_handling as ih
import tilesman.tests.helpers.file_system as fs
from tilesman.photo import *


TEST_DIR_PATH = os.path.join('.', 'test_tmp')
TEST_IMAGE_NAME = 'test_image'
TEST_IMAGE_FORMAT = 'jpg'
TEST_IMAGE_PATH = os.path.join(TEST_DIR_PATH, '.'.join([TEST_IMAGE_NAME, TEST_IMAGE_FORMAT]))
IMAGE_TILES_DIR_PATH = os.path.join(TEST_DIR_PATH, 'tiles', TEST_IMAGE_NAME)
IMAGE_TILING_LEVEL_NAME = 'tiling_level'

class TestPhotoMethods(unittest.TestCase):
	'''
	Unit tests for the photo module
	'''

	def setUp(self):
		fs.create_dir(TEST_DIR_PATH)
		ih.mock_image_file(TEST_IMAGE_PATH, 875, 625)

	def tearDown(self):
		fs.delete_dir(TEST_DIR_PATH)
	
	def test_photo_init_exception(self):
		with self.assertRaises(OSError):
			Photo(os.path.join(TEST_DIR_PATH, 'fake_test_image.jpg'))

	def test_photo_init(self):
		img = Photo(TEST_IMAGE_PATH)
		self.assertEqual(img.dir, TEST_DIR_PATH)
		self.assertEqual(img.name, TEST_IMAGE_NAME)
		self.assertEqual(img.file_format, TEST_IMAGE_FORMAT)
		self.assertEqual([img.width, img.height], [875, 625])

	def test_extract_tiles_exception(self):
		img = Photo(TEST_IMAGE_PATH)
		with self.assertRaises(ZeroDivisionError):
			img.extract_tiles(IMAGE_TILING_LEVEL_NAME, 0, 0)

	def test_extract_tiles(self):
		img = Photo(TEST_IMAGE_PATH)
		img.extract_tiles(IMAGE_TILING_LEVEL_NAME, 256, 256)

		# Get all extratec tiles by file name
		tile_dir_path = os.path.join(IMAGE_TILES_DIR_PATH, IMAGE_TILING_LEVEL_NAME)
		tile_names = fs.scan_dir(tile_dir_path, 'file')

		# Checking tile count
		self.assertEqual(len(tile_names), 12)
		
		# Check if all tiles in row y = 0 make up the same width of the original image
		self.assertEqual(
			sum([ih.load_image(os.path.join(tile_dir_path, tile_name)).size[0] for tile_name in tile_names if re.compile('\d_0\.jpg$').match(tile_name)]), 875
		)
		# Check if all tiles in column x = 0 make up the same height of the original image
		self.assertEqual(
			sum([ih.load_image(os.path.join(tile_dir_path, tile_name)).size[1] for tile_name in tile_names if re.compile('0_\d\.jpg$').match(tile_name)]), 625
		)

	def test_shrink_exception(self):
		img = Photo(TEST_IMAGE_PATH)
		with self.assertRaises(ValueError):
			img.shrink(-1)

	def test_shrink(self):
		img = Photo(TEST_IMAGE_PATH)
		img.shrink(0.5)
		self.assertEqual(img.width, 438)
		self.assertEqual(img.height, 313)



import os
from PIL import Image

def mock_generic_file(root_dir, name):
	'''
	Creates a generic mock file
	'''

	open(os.path.join(root_dir, name), "w").close()

def mock_image_on_memory(width, height, image_type = 'RGB', color = (100, 100, 100)):
	'''
	Creates a mock image. It is not saved to the file system
	'''

	return Image.new(image_type, size = (width, height), color = color)

def mock_image_file(file_path, width, height):
	'''
	Creates a mock image and saves it to the file system. File type is defined as per the extension
	provided in 'file_path'
	'''

	image = mock_image_on_memory(width, height)
	image.save(file_path)

def mock_several_image_files(root_dir, base_name, file_format, amount, width, height):
	'''
	Creates 'amount' mock images on 'root_dir', using 'base_name' as file name prefix. All images will have
	same dimensions ('width' x 'height' pixels), be of the same 'image_type' and 'color'
	'''

	for i in range(amount):
		file_path = os.path.join(root_dir, base_name + '_' + str(i) + '.' + file_format)
		mock_image_file(file_path, width, height)

def load_image(path):
	'''
	Loads on memory an image referenced by path
	'''
	
	return Image.open(path)




def photo_tiling(photo_path, tile_max_width, tile_max_height, ignore_input_file_exception = True):
	'''
	Given a photo in 'photo_path', will create a pyramid of tiles (maximum size of
	'tile_max_width' x 'tile_max_height' pixels) split into different photo resolution levels,
	represented in a folder hierarchy. If 'ignore_input_file_exception' is true, exceptions when
	trying to load files that are not images will be ignored.
	'''
	
	import os
	import sys
	sys.path.append(os.path.dirname(__file__))
	from math import log, ceil
	from photo import Photo

	try:
		img = Photo(photo_path)

		# Find expected number of resolution levels
		resolution_levels = 1 + ceil(log(max(img.width, img.height),2))

		# Iterate over resolution levels, from higuest to lowest
		for level in range(resolution_levels - 1, -1, -1):
			# Initiate the tiling process for the current level
			img.extract_tiles(level, tile_max_width, tile_max_height)

			# Resize photo to half the current size, if photo is larger than 1 x 1 pixels
			if img.width > 1 or img.height > 1:
				img.shrink(0.5)

	# The provided path does not reference a supported image type
	except OSError as e:
		if not ignore_input_file_exception:
			raise

def batch_photo_tiling(dir_path, tile_max_width, tile_max_height):
	'''
	Given a directory in 'dir_path', will orchestrate the tiling process for each photo in the
	provided directory, considering supplied tile maximum dimensions in pixels. This method uses
	parallel processing for performance improvement
	'''

	import os
	from multiprocessing import Pool 

	photo_paths = [os.path.join(dir_path, photo_name) for photo_name in next(os.walk(dir_path))[2]]
	params = [(photo_path, tile_max_width, tile_max_height) for photo_path in photo_paths]

	# Create a pool os processes (as many as the number of cores in the processing machine) and run 
	# tiling method in a multiprocessed fashion
	with Pool() as pool:
		pool.starmap(photo_tiling, params)

def prepare_argparse():
	import argparse

	argument_parser = argparse.ArgumentParser(
		description = 'Writes a pyramid of configurable tile sizes for one or more N x M pixels photos, at different resolution levels'
	)
	required_arg_group = argument_parser.add_argument_group('required arguments')
	required_arg_group.add_argument('-p', required = True, help = 'Path to a photo or a directory containing photos')
	required_arg_group.add_argument('-w', type = int, required = True, help = 'Maximum width for a tile')
	required_arg_group.add_argument('-e', type = int, required = True, help = 'Maximum height for a tile')
	return argument_parser

def main(args = None):
	'''
	Entry method for command line calls. Will trigger multiprocessed tiling if provided path
	is a directory. Single-processed tiling will be used otherwise (-p argument is a single file)
	'''

	import os

	argument_parser = prepare_argparse()
	args = argument_parser.parse_args()

	if os.path.isdir(args.p):
		batch_photo_tiling(args.p, args.w, args.e)
	else:
		photo_tiling(args.p, args.w, args.e)

if __name__ == '__main__':
	main()



import os
import unittest
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import tilesman.tests.helpers.image_handling as ih
import tilesman.tests.helpers.file_system as fs
from tilesman.tile import *


TEST_DIR = os.path.join('.', 'test_tmp')

class TestTileMethods(unittest.TestCase):
	'''
	Unit tests for the tile module
	'''

	def setUp(self):
		fs.create_dir(TEST_DIR)

	def tearDown(self):
		fs.delete_dir(TEST_DIR)

	def test_object_init(self):
		tile_binary = ih.mock_image_on_memory(400, 200)
		tile = Tile(tile_binary, (2, 4), 'jpg')
		self.assertEqual((tile.width, tile.height), (400, 200))
		self.assertEqual(tile.coordinates, (2, 4))
		self.assertEqual(tile.file_format, 'jpg')

	def test_save(self):
		tile_binary = ih.mock_image_on_memory(250, 500)
		tile = Tile(tile_binary, (2, 4), 'jpg')
		tile.save(TEST_DIR)

		expected_tile_path = os.path.join(TEST_DIR, '2_4.jpg')
		self.assertTrue(os.path.isfile(expected_tile_path))

		img = ih.load_image(expected_tile_path)
		self.assertEqual(img.size, (250, 500))

		os.remove(expected_tile_path)



from setuptools import setup

setup(
    name = 'tilesman',
    author = 'Fernando Urzedo',
    author_email = 'furzedo@yahoo.com',
    version = '0.2.0',
    packages = ['tilesman', 'tilesman.tests', 'tilesman.tests.helpers'],
    url = 'https://gitlab.com/furzedo/tilesman',
    license = 'LICENSE.txt',
    description = 'A library to tile photos at different resolution levels, for loading performance',
    long_description = open('README.txt').read(),
    python_requires = '>=3.3',
    install_requires = [
        "Pillow>=5.0.0",
    ],
    entry_points = {
        'console_scripts': [
            'tilesman = tilesman.core:main',
        ]
    }
)



import os
import unittest
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import tilesman.tests.helpers.image_handling as ih
import tilesman.tests.helpers.file_system as fs
from tilesman.core import photo_tiling, batch_photo_tiling


TEST_DIR_PATH = os.path.join('.', 'test_tmp')
TEST_IMAGE_NAME = 'test_image'
TEST_IMAGE_FORMAT = 'tif'
TEST_IMAGE_PATH = os.path.join(TEST_DIR_PATH, '.'.join([TEST_IMAGE_NAME, TEST_IMAGE_FORMAT]))
IMAGE_TILES_DIR_PATH = os.path.join(TEST_DIR_PATH, 'tiles', TEST_IMAGE_NAME)

class TestCore(unittest.TestCase):
	'''
	Integration tests for the tiling methods on module 'core'. Most of the tests will be triggered
	on the 'batch_photo_tiling' method, which consequently will also test method 'photo_tiling' as
	a by-product
	'''

	def setUp(self):
		fs.create_dir(TEST_DIR_PATH)

	def tearDown(self):
		fs.delete_dir(TEST_DIR_PATH)

	def __inspect_tile_dimensions(self, tile_inspection_scenarios):
		'''
		Verify if actual and expected tile dimensions match
		'''

		for level, (x, y), (width, height) in tile_inspection_scenarios:
			tile_path = os.path.join(IMAGE_TILES_DIR_PATH, str(level), '_'.join([str(x), str(y)]) + '.tif')
			self.assertEqual(ih.load_image(tile_path).size, (width, height))

	def test_raises_exception_for_non_image_file(self):
		ih.mock_image_file(TEST_IMAGE_PATH, 1, 1)
		ih.mock_generic_file(TEST_DIR_PATH, 'non_image_file.txt')

		with self.assertRaises(OSError):
			photo_tiling(os.path.join(TEST_DIR_PATH, 'non_image_file.txt'), 10, 10, ignore_input_file_exception = False)

	def test_tiling_empty_directory(self):
		batch_photo_tiling(TEST_DIR_PATH, 10, 10)

		# Checking directory structure and file count
		self.assertEqual(fs.recursive_file_count(TEST_DIR_PATH), 0)
		self.assertEqual(fs.recursive_dir_count(TEST_DIR_PATH), 0)

	def test_tiling_single_pixel_image(self):
		ih.mock_image_file(TEST_IMAGE_PATH, 1, 1)

		batch_photo_tiling(TEST_DIR_PATH, 5, 10)

		# Checking directory structure and file count
		self.assertEqual(sorted(fs.scan_dir(IMAGE_TILES_DIR_PATH)), ['0'])
		self.assertEqual(fs.recursive_file_count(os.path.join(IMAGE_TILES_DIR_PATH, '0')), 1)

		# Checking some tiles at critical positions
		# scenario = (level, (x_index, y_index), (width, height))
		tile_inspection_scenarios = [(0, (0, 0), (1, 1)),] # the only tile produced at level 0
		self.__inspect_tile_dimensions(tile_inspection_scenarios)

	def test_tiling_regular_image(self):
		ih.mock_image_file(TEST_IMAGE_PATH, 7000, 5000)

		batch_photo_tiling(TEST_DIR_PATH, 256, 256)

		# Checking directory structure and file count
		self.assertEqual(sorted(fs.scan_dir(IMAGE_TILES_DIR_PATH)), ['0', '1', '10', '11', '12', '13', '2', '3', '4', '5', '6', '7', '8', '9'])
		self.assertEqual(fs.recursive_file_count(os.path.join(IMAGE_TILES_DIR_PATH, '0')), 1)
		self.assertEqual(fs.recursive_file_count(os.path.join(IMAGE_TILES_DIR_PATH, '10')), 12)
		self.assertEqual(fs.recursive_file_count(os.path.join(IMAGE_TILES_DIR_PATH, '13')), 560)

		# Checking some tiles at critical positions
		# scenario = (level, (x_index, y_index), (width, height))
		tile_inspection_scenarios = [
			(0, (0, 0), (1, 1)), # the only tile produced at level 0
			(10, (1, 1), (256, 256)), # a full-dimensioned tile at level 10
			(10, (3, 0), (107, 256)), # a level 10 tile at the edge of x axis
			(10, (0, 2), (256, 113)), # a level 10 tile at the edge of y axis
			(10, (3, 2), (107, 113)), # the level 10 tile at the edge of x and y axes
			(13, (15, 15), (256, 256)), # a full-dimensioned tile at level 13
			(13, (27, 5), (88, 256)), # a level 13 tile at the edge of y axis
			(13, (10, 19), (256, 136)), # a level 13 tile at the edge of y axis
			(13, (27, 19), (88, 136)), # the level 13 tile at the edge of x and y axes
		]
		self.__inspect_tile_dimensions(tile_inspection_scenarios)

	def test_tiling_several_images_in_a_directory(self):
		amount_of_images = 5

		ih.mock_several_image_files(TEST_DIR_PATH, TEST_IMAGE_NAME, TEST_IMAGE_FORMAT, amount_of_images, 1000, 800)
		ih.mock_generic_file(TEST_DIR_PATH, 'non_image_file.txt')

		batch_photo_tiling(TEST_DIR_PATH, 150, 100)

		# Checking that the generic file did not influence the tiling process
		self.assertFalse(os.path.exists(os.path.join(TEST_DIR_PATH, 'tiles', 'non_image_file')))

		for i in range(amount_of_images):
			image_i_tiles_dir_path = IMAGE_TILES_DIR_PATH + '_' + str(i)

			# Checking directory structure and file count
			self.assertEqual(sorted(fs.scan_dir(image_i_tiles_dir_path)), ['0', '1', '10', '2', '3', '4', '5', '6', '7', '8', '9'])
			self.assertEqual(fs.recursive_file_count(os.path.join(image_i_tiles_dir_path, '0')), 1)
			self.assertEqual(fs.recursive_file_count(os.path.join(image_i_tiles_dir_path, '4')), 1)
			self.assertEqual(fs.recursive_file_count(os.path.join(image_i_tiles_dir_path, '8')), 4)
			self.assertEqual(fs.recursive_file_count(os.path.join(image_i_tiles_dir_path, '10')), 56)
